"""
This module will demo the dt_tools.console.console_helper capabilities including:

- cursor positioning (move to, print at, ...)
- console clearing (screen, line, to EOL, to BOL,...)
- setting FG and BG colors, wrapping words with color, ...

"""
from dt_tools.console.console_helper import ColorBG, ColorFG, TextStyle, _CursorAttribute, CursorShape
from dt_tools.console.console_helper import ConsoleHelper as console
from dt_tools.console.console_helper import ConsoleInputHelper as console_input
from dt_tools.os.os_helper import OSHelper
import time

def demo():
    OSHelper.enable_ctrl_c_handler()
    wait_seconds = 2

    console.cursor_set_shape(CursorShape.BLINKING_UNDERLINE)
    console.clear_screen(cursor_home=True)
    row, col = console.cursor_current_position()
    console_size = console.get_console_size()
    console.set_console_viewport(1,console_size[0]-1)
    console.print_with_wait(f'Console size: {console_size}, cur pos: {row},{col}', wait_seconds, eol='\n\n')
    console.cursor_save_position()

    console.print_line_separator('Test color attributes', 40)
    token = console.cwrap('string', fg=ColorFG.RED, bg=ColorBG.WHITE, style=TextStyle.ITALIC)
    console.print(f'This {token} is Red Italic on White BG')
    console.print(f'RAW: This {token} is Red Italic on White BG', as_bytes=True)
    print(f'This {console.cwrap("string", ColorFG.GREEN)} is Green')
    print(f'This {console.cwrap("string", ColorFG.RED)} is Red')
    print(f'This {console.cwrap("string", ColorFG.RED, None, TextStyle.ITALIC)} is Italic Red')
    print(f'This {console.cwrap("string", ColorFG.RED, None, TextStyle.BOLD)} is Bold Red\n')
    console_input.get_input_with_timeout('Press ENTER to continue ', timeout_secs=10)
    
    console.cursor_restore_position()
    console.clear_to_EOS()

    console.print_line_separator('Test cursor attributes', 40)
    for attr in _CursorAttribute:
        console.debug_display_cursor_location()
        console.print_with_wait(f'CURSOR: {attr}', eol=' ')
        console.cursor_set_attribute(attr)
        console.print_with_wait('', wait_seconds, eol='')
        print('')
    print('')
    
    console_input.get_input_with_timeout('Press ENTER to continue ', timeout_secs=10)
    console.cursor_restore_position()
    console.clear_to_EOS()

    console.print_line_separator('Test cursor shape...', 40)
    for shape in CursorShape:
        console.debug_display_cursor_location()
        console.cursor_set_shape(shape)
        console.print(f'CURSOR: {shape}', eol= ' ')
        console.print_with_wait('', wait_seconds, eol='')
        # console.print_with_wait(f'CURSOR: {shape} {shape.value}', wait_seconds, eol = ' ')
        print('')
    
    console_input.get_input_with_timeout('Press ENTER to continue ', timeout_secs=10)
    console.clear_screen()

    console.cursor_set_shape(CursorShape.STEADY_BLOCK)            
    console.display_status('Test Rows...')
    for row in range(1, console_size[0]):
        console.print_at(row, 60, f'Row {row}', eol='')
    console.cursor_move(row=1,column=1)
    console.print_with_wait(f'Console size: {console_size} and current position: {row},{col}', wait_seconds)
    console.cursor_move(5,1)
    print(f'Look at the beautiful {console.cwrap("blue",ColorFG.BLUE2, style=TextStyle.BOLD)} sky')
    console.debug_display_cursor_location(f'After {console.cwrap("blue",ColorFG.BLUE2, style=TextStyle.BOLD)} sky')
    time.sleep(wait_seconds)

    print('Check cursor positioning...')
    console.print_at(10, 5, "Should print at  location 10,5 xxxxxxx", eol='')
    console.debug_display_cursor_location()
    time.sleep(wait_seconds)
    console.cursor_left(7)

    console.clear_to_EOL()
    console.debug_display_cursor_location(f"Clear to {console.cwrap('EOL',ColorFG.GREEN2, style=TextStyle.BOLD)}")
    time.sleep(wait_seconds)

    print('abc', end='')
    console.debug_display_cursor_location()
    time.sleep(wait_seconds)

    console.clear_to_BOL()
    console.debug_display_cursor_location(f"Clear to {console.cwrap('BOL',ColorFG.GREEN2, style=TextStyle.BOLD)}")
    time.sleep(wait_seconds)

    console.clear_to_BOS()
    console.debug_display_cursor_location(f"Clear to {console.cwrap('BOS',ColorFG.GREEN2, style=TextStyle.BOLD)}")
    time.sleep(wait_seconds)

    console.cursor_move(12,1)
    console.debug_display_cursor_location( "Moved to 12,1")
    time.sleep(wait_seconds)

    console.clear_to_EOS()
    console.debug_display_cursor_location(f"Clear to {console.cwrap('EOS',ColorFG.GREEN2, style=TextStyle.BOLD)}")
    time.sleep(wait_seconds)

    console.print_with_wait(f'Console size: {console_size}, cur pos: {row},{col}', wait_seconds, eol='\n\n')

    console.clear_screen()
    console.set_console_viewport(start_row=2, end_row=console_size[0]-1)
    console.print_line_separator('Check scrolling...', 40)
    for row in range(1, 50):
        console.print(f'Row {row}')    
        if row % 5 == 0:
            console.debug_display_cursor_location('Scrolling...')
            time.sleep(.5)
    console_input.get_input_with_timeout('Press ENTER to continue ', timeout_secs=10)

    console.set_console_viewport()
    console.clear_screen()
    console.print_line_separator('Display color palette, codes are [style,fg,bg]...', 40)
    console.print('')
    time.sleep(wait_seconds)
    console._display_color_palette()

    console.cursor_set_shape(CursorShape.DEFAULT)
    print(f"End of {console.cwrap('ConsoleHelper', ColorFG.YELLOW)} demo.")

if __name__ == '__main__':
    demo()


"""
Driver for the Console Demos.

This module will walk you thru each dt_tools.console control and demo
the capabilities.

- Console manipulation: console_helper :func:`~dt_tools.cli.dt_console_helper_demo.demo()`
- Console input: console_input_helper :func:`~dt_tools.cli.dt_console_input_helper_demo.demo()`
- GUI Messagebox: msgbox :func:`~dt_tools.cli.msgbox_demo.demo()`
- Progress Bar control: progress_bar :func:`~dt_tools.cli.dt_progress_bar_demo.demo()`
- Spinner control: spinner :func:`~dt_tools.cli.dt_spinner_demo.demo()`

To Run:
    `poetry run python -m dt_tools.cli.dt_console_demo`

"""
import dt_tools.logger.logging_helper as lh
from dt_tools.cli.demos.dt_console_helper_demo import demo as console_helper_demo
from dt_tools.cli.demos.dt_console_input_helper_demo import demo as console_input_helper_demo
from dt_tools.cli.demos.dt_msgbox_demo import demo as message_box_demo
from dt_tools.cli.demos.dt_progress_bar_demo import demo as progress_bar_demo
from dt_tools.cli.demos.dt_spinner_demo import demo as spinner_demo
from dt_tools.console.console_helper import ColorFG, TextStyle
from dt_tools.console.console_helper import ConsoleHelper as console
from dt_tools.console.console_helper import ConsoleInputHelper as console_input
from dt_tools.os.project_helper import ProjectHelper

def run_demos():
    lh.configure_logger(log_level="INFO")
    DEMOS = {
        "ConsoleHelper": console_helper_demo,
        "ConsoleInputHelper": console_input_helper_demo,
        "MessageBox": message_box_demo,
        "ProgressBar": progress_bar_demo,
        "Spinner": spinner_demo
    }
    console.clear_screen()
    console.print_line_separator('', 80)
    version = f'v{console.cwrap(ProjectHelper.determine_version("dt-console"), style=TextStyle.ITALIC)}'
    console.print_line_separator(f'dt_console_demo {version}', 80)
    console.print('')
    for name, demo_func in DEMOS.items():
        demo_name = console.cwrap(name, ColorFG.YELLOW)
        resp = console_input.get_input_with_timeout(f'Demo {demo_name} Functions (y/n) > ', 
                                                console_input.YES_NO_RESPONSE, default='n', 
                                                timeout_secs=10).lower()
        if resp == 'y':
            demo_func()
            console.print('')

    console.print('Console demo complete!')
    
if __name__ == '__main__':
    run_demos()


"""
Spinner control to provide visual feedback of process activity.

Features:
   - Customizable display.    
   - Optionally, display elapsed time.
   - Selectable spinner icons.

Example::
    from dt_tools.console.spinner import Spinner, SpinnerType

    spinner = Spinner("In progress", SpinnerType.DOTS, True)
    spinner.start_spinner()
    # Do some work....
    spinner.stop_spinner()
    
"""
import threading
import time
from datetime import datetime as dt
from enum import Enum

from loguru import logger as LOGGER
from dt_tools.console.console_helper import ConsoleHelper


class SpinnerType(Enum):
    """Spinner types, used in Spinner constructor"""
    NORMAL_SPINNER={"speed": .075, "char_list": ['|','/','-','\\']}
    ARC = {"speed": .075, "char_list":  ["◜","◠","◝","◞","◡","◟"]}
    BALL_BOUNCER = {"speed": .1, "char_list": ["( ●    )","(  ●   )","(   ●  )","(    ● )","(     ●)","(    ● )","(   ●  )","(  ●   )","( ●    )","(●     )"]}
    BALLOON = {"speed": .1, "char_list": [".","o","O","°","O","o","."]}
    BLOCK_BLINK={"speed": .25, "char_list": [' ','▓', '█']}
    BLOCK_BOUNCER = {"speed": .1, "char_list": ["▐⠂       ▌","▐⠈       ▌","▐ ⠂      ▌","▐ ⠠      ▌","▐  ⡀     ▌","▐  ⠠     ▌","▐   ⠂    ▌","▐   ⠈    ▌","▐    ⠂   ▌","▐    ⠠   ▌","▐     ⡀  ▌","▐     ⠠  ▌","▐      ⠂ ▌","▐      ⠈ ▌","▐       ⠂▌","▐       ⠠▌","▐       ⡀▌","▐      ⠠ ▌","▐      ⠂ ▌","▐     ⠈  ▌","▐     ⠂  ▌","▐    ⠠   ▌","▐    ⡀   ▌","▐   ⠠    ▌","▐   ⠂    ▌","▐  ⠈     ▌","▐  ⠂     ▌","▐ ⠠      ▌","▐ ⡀      ▌","▐⠠       ▌"]   }
    BLOCK_GROW_UP = {"speed": .1, "char_list": ["▁","▃","▄","▅","▆","▇","▆","▅","▄","▃"]}
    CIRCLE_SPIN = {"speed": .1,"char_list":["◐","◓","◑","◒"]}
    DOTS = {"speed": .1, "char_list": ["⠋","⠙","⠹","⠸","⠼","⠴","⠦","⠧","⠇","⠏"]}
    FADING_BOX = {"speed": .1, "char_list": ['█','▓','▒','░',' ']}
    FLIPPER = {"speed": .075, "char_list": ["_","_","_","-","`","`","'","´","-","_","_","_"] }
    MOON_PHASES = {"speed": .15, "char_list": ["🌑 ","🌒 ","🌓 ","🌔 ","🌕 ","🌖 ","🌗 ","🌘 "]}

class Spinner():
    """
    Create a console spinner for visual effect.  
    
    The spinner runs on a seperate thread, so the caller can perform 
    processing while the spinner displays.
    
    Example::
    
        from dt_tools.console.spinner import Spinner, SpinnerType

        spinner = Spinner("In progress", SpinnerType.DOTS, True)
        spinner.start_spinner()
        # Do some work....
        spinner.stop_spinner()
        
    Parameters:
        caption     : string prefixing spinner graphic
        spinner     : type of spinner pattern
        show_elapsed: suffix displaying elapsed h:m:s
    """
    def __init__(self, caption: str, spinner: SpinnerType = SpinnerType.NORMAL_SPINNER, show_elapsed: bool = False, str_end = ''):
        self._caption = caption
        self._suffix = ''
        self._last_suffix = ''
        self._spinner = spinner
        self._cursor_list = self._spinner.value['char_list']     
        self._cursor_list_len = len(self._cursor_list)
        self._show_elapsed = show_elapsed
        self._elapsed_time = '00:00:00'
        self._start_time = dt.now()
        self._str_end = str_end
        self._idx = 99
        self._finished = False
        self._spinner_thread = None
        # self.console = ConsoleHelper()
        LOGGER.trace("Spinner initialized.")

    def start_spinner(self, caption_suffix: str = ''):
        """
        Start spinner in console.

        Keyword Arguments:
            caption_suffix:  Text to append to spinner line (default: {''})
        """
        if self._spinner_thread:
            # If spinner is currently running, kill it.
            if self._spinner_thread.is_alive():
                self.stop_spinner()

        ConsoleHelper.cursor_off()
        if ConsoleHelper.valid_console():
            self._suffix = caption_suffix
            self._spinner_thread = threading.Thread(target=self._display_spinner, daemon=True)
            self._start_time = dt.now()
            self._elapsed_time = '00:00:00'
            self._spinner_thread.start()
    
    def stop_spinner(self):
        """
        Stop spinner when running

        Spinner line will be cleared and cursor will be positioned 
        in column 1 of that row
        """
        if self._spinner_thread is not None and self._spinner_thread.is_alive():
            self._finished = True
            self._spinner_thread.join()
            self._elapsed_time = self._calculate_elapsed_time(dt.now(), self._start_time)
            self._finished = False
        ConsoleHelper.clear_line()
        ConsoleHelper.cursor_on()

    def caption_suffix(self, suffix: str):
        """
        Text to append at end of spinner line.

        Use this to provide updated status text as the spinner is running.

        Arguments:
            caption_suffix:  Text to append to spinner line
        """
        self._suffix = suffix

    @property
    def elapsed_time(self) -> str:
        """
        Elapsed time since spinner started.

        Returns:
            String in format hh:mm:ss
        """
        return self._elapsed_time


    def _calculate_elapsed_time(self, end_time: float, start_time: float) -> str:
        diff = end_time - start_time
        hours = diff.seconds // 3600
        minutes = diff.seconds // 60 % 60
        seconds = diff.seconds % 60
        elapsed_time = f"{hours:02d}h:{minutes:02d}m:{seconds:02d}s"
        return elapsed_time

    def _display_spinner(self):
        delay = self._spinner.value['speed']
        loopcnt = 0
        elapsed_break = int(1 / delay) # 1 seconds
        elapsed_display = ' '*len(self._elapsed_time)
        while not self._finished and threading.main_thread().is_alive():
            cursor = self._get_cursor()
            # suffix = self._calculate_suffix()

            if self._show_elapsed and loopcnt % elapsed_break == 0:
                self._elapsed_time = self._calculate_elapsed_time(dt.now(), self._start_time) # type: ignore
                elapsed_display = self._elapsed_time
            while not ConsoleHelper.cursor_move(column=1):
                # print('oops')
                time.sleep(.01)
            # time.sleep(.05)
            terminal_line = f'{self._caption} {cursor}  {elapsed_display} {self._suffix}'
            ConsoleHelper.print(terminal_line, eol='')
            # time.sleep(.05)
            ConsoleHelper.clear_to_EOL()
            time.sleep(delay)
            loopcnt += 1

        # terminal_line = ' ' * len(terminal_line)
        ConsoleHelper.clear_line()
        # print(terminal_line, end = self._str_end)

    def _get_cursor(self):
        self._idx += 1
        if self._idx >= self._cursor_list_len:
            self._idx = 0
        return self._cursor_list[self._idx]


if __name__ == "__main__":
    for spinner_type in SpinnerType:
        spinner = Spinner(f'Demo of spinner type: {spinner_type.name}', spinner_type, True)
        spinner.start_spinner("Begin")
        for cnt in range(1,25):
            if cnt % 5 == 0:
                spinner.caption_suffix(f'Executing iteration {cnt}')
            time.sleep(.25)
        spinner.stop_spinner()



"""
This module will demo the dt_tools.console.input_helper capabilities including:

- Prompt with timeout and default response
- Wait with Timout prompt
- Validate input prompt

To run this demo standalone:
    `poetry run python -m dt_tools.cli.dt_console_input_helper_demo`

"""
from dt_tools.console.console_helper import ColorFG, TextStyle
from dt_tools.console.console_helper import ConsoleHelper as console
from dt_tools.console.console_helper import ConsoleInputHelper as console_input
from dt_tools.os.os_helper import OSHelper


def demo():    
    OSHelper.enable_ctrl_c_handler()
    timeout = 10
    console.print('')
    console.print_line_separator('ConsoleInputHelper Demo', 40)
    
    console.print('')
    test_name = console.cwrap('Input with Timeout', style=[TextStyle.ITALIC, TextStyle.BOLD])    
    console.print(f'{test_name}: default response is y, timeout {timeout} secs...')
    console.print(console.cwrap("code:\n  resp = console_input.get_input_with_timeout('Test prompt (y/n) > ', console_input.YES_NO_RESPONSE, default='y', timeout_secs=timeout)\n",TextStyle.ITALIC))
    resp = console_input.get_input_with_timeout('Test prompt (y/n) > ', console_input.YES_NO_RESPONSE, default='y', timeout_secs=timeout)
    console.print(f'  returns: {resp}')

    console.print("")
    test_name = console.cwrap('Wait with Timeout', style=[TextStyle.ITALIC, TextStyle.BOLD] )
    console.print(f'\n{test_name}: Wait with bypass -  {timeout} seconds')
    console.print(console.cwrap("code:\n  resp = console_input.wait_with_bypass(timeout)\n",TextStyle.ITALIC))
    console.print(f'Waiting {timeout} seconds, or press ENTER to abort wait', eol='')
    if console_input.wait_with_bypass(timeout):
        console.print('\n  Prompt timed out.')
    else:
        console.print('  User aborted wait')

    console.print('')
    test_name = console.cwrap('Validate input (with timeout)', style=[TextStyle.ITALIC, TextStyle.BOLD])
    console.print(f'\n{test_name}: Wait {timeout} seconds, validate user enters proper input, if timeout return default value.')
    console.print(console.cwrap("code:\n  resp = console_input.get_input_with_timeout('Valid input is [a,b,c,d] > ', valid_responses=['a','b','c','d'], default='z', timeout_secs=timeout)\n",TextStyle.ITALIC))
    resp = console_input.get_input_with_timeout('Valid input is [a,b,c,d] > ', valid_responses=['a','b','c','d'], default='z', timeout_secs=timeout)
    print(f'  returns: {resp}')
    
    console.print('')
    console.print(f"End of {console.cwrap('ConsoleInputHelper', ColorFG.YELLOW)} demo.")

    input('\nPress Enter to continue...')

if __name__ == '__main__':
    demo()


"""
Helper to display tkinter Message box on display

Message boxes provided:

    - alert: Alert message box with message text and single OK button.
    - confirm: Confirmation message box with OK/Cancel buttons.  Returns caption of button clicked.
    - prompt: Display message box with text input and OK/Cancel buttons.  Returns text entered, or None if Cancel clicked.
    - password: Displays a passworkd message box and OK/Cancel buttons.  Input is masked. Returns text entered, or None if Cancel clicked.

Features:
    - Ability to set Timeout seconds.  Message box will return 'timeout' when triggered.
    - Confirm message box allows user to customize buttons available.

"""
import tkinter as tk
from typing import List, Tuple, Union
from enum import Enum
from loguru import logger as LOGGER

# This version derived from the below projects:

# Modified BSD License
# PyMsgBox - A simple, cross-platform, pure Python module for JavaScript-like message boxes.
#            By Al Sweigart al@inventwithpython.com

__version__ = "1.0.9"

# Modified BSD License
# Derived from Stephen Raymond Ferg's EasyGui http://easygui.sourceforge.net/
"""
The four functions in PyMsgBox:

 - alert(text='', title='', button='OK')

    Displays a simple message box with text and a single OK button. Returns the text of the button clicked on.

 - confirm(text='', title='', buttons=['OK', 'Cancel'])

    Displays a message box with OK and Cancel buttons. Number and text of buttons can be customized. Returns the text of the button clicked on.

 - prompt(text='', title='' , default='')

    Displays a message box with text input, and OK & Cancel buttons. Returns the text entered, or None if Cancel was clicked.

 - password(text='', title='', default='', mask='*')

    Displays a message box with text input, and OK & Cancel buttons. Typed characters appear as *. Returns the text entered, or None if Cancel was clicked.
"""

"""
TODO Roadmap:
- Be able to specify a custom icon in the message box.
- Be able to place the message box at an arbitrary position (including on multi screen layouts)
- Add mouse clicks to unit testing.
- progress() function to display a progress bar
- Maybe other types of dialog: open, save, file/folder picker, etc.
"""

rootWindowPosition = "+300+200"

class MB_FontFamily:
    """MsgBox Font Types."""
    PROPORTIONAL = ("MS", "Sans", "Serif")
    MONOSPACE = "Courier"

class MB_FontSize(Enum):
    """Font sizes for MsgBox fonts"""
    DEFAULT:int = -1
    PROPORTIONAL:int = 10
    MONOSPACE:int = 9  # a little smaller, because it it more legible at a smaller size
    TEXT:int = 12        # a little larger makes it easier to see


MSGBOX_WIDTH = 400

STANDARD_SELECTION_EVENTS = ["Return", "Button-1", "space"]

# constants for strings: (TODO: for internationalization, change these)
class MB_ButtonType:
    """Constants for Message box buttons (and captions)."""
    OK = "OK"
    CANCEL = "Cancel"
    YES = "Yes"
    NO = "No"
    RETRY = "Retry"
    ABORT = "Abort"
    IGNORE = "Ignore"
    TRY_AGAIN = "Try Again"
    CONTINUE = "Continue"

TIMEOUT_RETURN_VALUE = "Timeout"
"""Value returned when MsgBox times out."""

# Initialize some global variables that will be reset later
_used_font_family = MB_FontFamily.PROPORTIONAL
_used_font_size = MB_FontSize.PROPORTIONAL.value
__choiceboxMultipleSelect = None
__widgetTexts = None
__replyButtonText = None
__choiceboxResults = None
__firstWidget = None
__enterboxText = None
__enterboxDefaultText = ""
__multenterboxText = ""
choiceboxChoices = None
choiceboxWidget = None
entryWidget = None
boxRoot = None
buttonsFrame = None


def set_font(family: MB_FontFamily, size: MB_FontSize = MB_FontSize.DEFAULT):
    global _used_font_family, _used_font_size
    """Set font family and font size

    Arguments:
        family: MB_FontFamily.MONOSPACE or MB_FontFamily.PROPORTIONAL.
        size: Font size, typically 8-12.  If <= 0 will default to
          either MB_FontSize.MONOSPACE or MB_FontSize.PROPORTIONAL depending
          on select family. 
    """
    _used_font_family = family
    if size.value <= MB_FontSize.DEFAULT.value:
        _used_font_size = MB_FontSize.MONOSPACE.value if _used_font_family == MB_FontFamily.MONOSPACE else MB_FontSize.PROPORTIONAL.value
    else:
        _used_font_size = size.value


def _alertTkinter(text="", title="", button=MB_ButtonType.OK, root=None, timeout=None):
    """
    Displays a simple message box with text and a single OK button. 

    Keyword Arguments:
        text: Text content of the message box (default: {""})
        title: Title bar text (default: {""})
        button: Button caption (default: {MB_ButtonType.OK})
        root: Base window (default: {None})
        timeout: Number of seconds to display box (default: {None})

    Returns:
        The text of the button clicked on.
    """
    text = str(text)
    retVal = _buttonbox(
        msg=text, title=title, choices=[str(button)], root=root, timeout=timeout
    )
    return button if retVal is None else retVal

alert = _alertTkinter


def _confirmTkinter(
    text="", title="", buttons=(MB_ButtonType.OK, MB_ButtonType.CANCEL), root=None, timeout=None):
    """
    Displays a Confirmation message box with text and Ok / Cancel buttons. 

    Keyword Arguments:
        text: Text content of the message box (default: {""})
        title: Title bar text (default: {""})
        buttons: A list of button text (default: {(MB_ButtonType.OK, MB_ButtonType.CANCEL)})
        root: Base window (default: {None})
        timeout: Number of seconds to display box (default: {None})

    Returns:
        The text of the button clicked on.
    """

    text = str(text)
    resp = _buttonbox(
        msg=text,
        title=title,
        choices=[str(b) for b in buttons],
        root=root,
        timeout=timeout,
    )

    return resp

confirm = _confirmTkinter


def _promptTkinter(text="", title="", default="", root=None, timeout=None):
    """
    Displays a message box with text input, and OK & Cancel buttons. 

    Keyword Arguments:
        text: Text content of the message box (default: {""})
        title: Title bar text (default: {""})
        default: Default text on timeout (default: {""})
        root: Base window (default: {None})
        timeout: Number of seconds to display box (default: {None})

    Returns:
        Returns the text entered, or None if Cancel was clicked.
    """
    text = str(text)
    return __fillablebox(
        text, title, default=default, mask=None, root=root, timeout=timeout
    )


prompt = _promptTkinter


def _passwordTkinter(text="", title="", default="", mask="*", root=None, timeout=None):
    """
    Displays a message box with text input, and OK & Cancel buttons. 
    
    By default, typed characters are masked as an asterisk(\*) symbol. 

    Keyword Arguments:
        text: Text content of the message box (default: {""})
        title: Title bar text (default: {""})
        mask: Mask character to display on user keystrokes (default: {"*"})
        default: Default text on timeout (default: {""})
        root: Base window (default: {None})
        timeout: Number of seconds to display box (default: {None})

    Returns:
        Returns the text entered, or None if Cancel was clicked.
    """
    text = str(text)
    return __fillablebox(text, title, default, mask=mask, root=root, timeout=timeout)

password = _passwordTkinter


def timeoutBoxRoot():
    global boxRoot, __replyButtonText, __enterboxText
    try:
        boxRoot.destroy() # timeoutBoxRoot
    except Exception as ex:
        LOGGER.trace(f'timeoutBoxRoot()-{repr(ex)}')

    __replyButtonText = TIMEOUT_RETURN_VALUE
    __enterboxText = TIMEOUT_RETURN_VALUE


def _buttonbox(msg: str, title: str, choices: Union[List,Tuple], root=None, timeout=None):
    """
    Display a msg, a title, and a set of buttons.

    The buttons are defined by the members of the choices list.

    Arguments:
        msg: Text content of the message box
        title: Title bar text
        choices: A list or tuple of the choices to be displayed

    Keyword Arguments:
        root: Base window (default: {None})
        timeout: Number of seconds to display box (default: {None})

    Returns:
        Text caption of the Button clicked.
    """

    global boxRoot, __replyButtonText, __widgetTexts, buttonsFrame

    # Initialize __replyButtonText to the first choice.
    # This is what will be used if the window is closed by the close button.
    __replyButtonText = choices[0]

    if root:
        root.withdraw()
        boxRoot = tk.Toplevel(master=root)
        boxRoot.withdraw()
    else:
        boxRoot = tk.Tk()
        boxRoot.withdraw()

    boxRoot.title(title)
    boxRoot.iconname("Dialog")
    boxRoot.geometry(rootWindowPosition)
    boxRoot.minsize(400, 100)

    # ------------- define the messageFrame ---------------------------------
    messageFrame = tk.Frame(master=boxRoot)
    messageFrame.pack(side=tk.TOP, fill=tk.BOTH)

    # ------------- define the buttonsFrame ---------------------------------
    buttonsFrame = tk.Frame(master=boxRoot)
    buttonsFrame.pack(side=tk.TOP, fill=tk.BOTH)

    # -------------------- place the widgets in the frames -----------------------
    messageWidget = tk.Message(messageFrame, text=msg, width=MSGBOX_WIDTH)
    # messageWidget.configure(font=(MB_FontFamily.PROPORTIONAL, MB_FontSize.PROPORTIONAL))
    messageWidget.configure(font=(_used_font_family, _used_font_size))
    messageWidget.pack(side=tk.TOP, expand=tk.YES, fill=tk.X, padx="3m", pady="3m")

    __put_buttons_in_buttonframe(choices)

    # -------------- the action begins -----------
    # put the focus on the first button
    __firstWidget.focus_force()
    boxRoot.attributes("-topmost", True)  # agd
    boxRoot.deiconify()
    _timeout_id: str = None
    if timeout is not None:
        _timeout_id = boxRoot.after(timeout, timeoutBoxRoot)
    boxRoot.mainloop()

    if _timeout_id is not None:
        boxRoot.after_cancel(_timeout_id)

    try:
        boxRoot.destroy()
    except tk.TclError:
        if __replyButtonText != TIMEOUT_RETURN_VALUE:
            __replyButtonText = None

    if root:
        root.deiconify()
    return __replyButtonText


def __put_buttons_in_buttonframe(choices):
    """Put the buttons in the buttons frame"""
    global __widgetTexts, __firstWidget, buttonsFrame

    __firstWidget = None
    __widgetTexts = {}

    i = 0

    for buttonText in choices:
        tempButton = tk.Button(buttonsFrame, takefocus=1, text=buttonText)
        _bindArrows(tempButton)
        tempButton.pack(
            expand=tk.YES, side=tk.LEFT, padx="1m", pady="1m", ipadx="2m", ipady="1m"
        )

        # remember the text associated with this widget
        __widgetTexts[tempButton] = buttonText

        # remember the first widget, so we can put the focus there
        if i == 0:
            __firstWidget = tempButton
            i = 1

        # for the commandButton, bind activation events to the activation event handler
        commandButton = tempButton
        handler = __buttonEvent
        for selectionEvent in STANDARD_SELECTION_EVENTS:
            commandButton.bind("<%s>" % selectionEvent, handler)

        if MB_ButtonType.CANCEL in choices:
            commandButton.bind("<Escape>", __cancelButtonEvent)


def _bindArrows(widget, skipArrowKeys=False):
    widget.bind("<Down>", _tabRight)
    widget.bind("<Up>", _tabLeft)

    if not skipArrowKeys:
        widget.bind("<Right>", _tabRight)
        widget.bind("<Left>", _tabLeft)


def _tabRight(event):
    boxRoot.event_generate("<Tab>")


def _tabLeft(event):
    boxRoot.event_generate("<Shift-Tab>")


def __buttonEvent(event):
    """
    Handle an event that is generated by a person clicking a button.
    """
    global boxRoot, __widgetTexts, __replyButtonText
    __replyButtonText = __widgetTexts[event.widget]
    boxRoot.quit()  # quit the main loop


def __cancelButtonEvent(event):
    """Handle pressing Esc by clicking the Cancel button."""
    global boxRoot, __widgetTexts, __replyButtonText
    __replyButtonText = MB_ButtonType.CANCEL
    boxRoot.quit()


def __fillablebox(msg, title="", default="", mask=None, root=None, timeout=None):
    """
    Show a box in which a user can enter some text.
    You may optionally specify some default text, which will appear in the
    enterbox when it is displayed.
    Returns the text that the user entered, or None if he cancels the operation.
    """

    global boxRoot, __enterboxText, __enterboxDefaultText
    global cancelButton, entryWidget, okButton

    if title is None:
        title == ""
    if default is None:
        default = ""
    __enterboxDefaultText = default
    __enterboxText = __enterboxDefaultText

    if root:
        root.withdraw()
        boxRoot = tk.Toplevel(master=root)
        boxRoot.withdraw()
    else:
        boxRoot = tk.Tk()
        boxRoot.withdraw()

    boxRoot.title(title)
    boxRoot.iconname("Dialog")
    boxRoot.geometry(rootWindowPosition)
    boxRoot.bind("<Escape>", __enterboxCancel)

    # ------------- define the messageFrame ---------------------------------
    messageFrame = tk.Frame(master=boxRoot)
    messageFrame.pack(side=tk.TOP, fill=tk.BOTH)

    # ------------- define the buttonsFrame ---------------------------------
    buttonsFrame = tk.Frame(master=boxRoot)
    buttonsFrame.pack(side=tk.TOP, fill=tk.BOTH)

    # ------------- define the entryFrame ---------------------------------
    entryFrame = tk.Frame(master=boxRoot)
    entryFrame.pack(side=tk.TOP, fill=tk.BOTH)

    # ------------- define the buttonsFrame ---------------------------------
    buttonsFrame = tk.Frame(master=boxRoot)
    buttonsFrame.pack(side=tk.TOP, fill=tk.BOTH)

    # -------------------- the msg widget ----------------------------
    messageWidget = tk.Message(messageFrame, width="4.5i", text=msg)
    # messageWidget.configure(font=(MB_FontFamily.PROPORTIONAL, MB_FontSize.PROPORTIONAL))
    messageWidget.configure(font=(_used_font_family, _used_font_size))
    messageWidget.pack(side=tk.RIGHT, expand=1, fill=tk.BOTH, padx="3m", pady="3m")

    # --------- entryWidget ----------------------------------------------
    entryWidget = tk.Entry(entryFrame, width=40)
    _bindArrows(entryWidget, skipArrowKeys=True)
    # entryWidget.configure(font=(MB_FontFamily.PROPORTIONAL, TEXT_ENTRY_FONT_SIZE))
    entryWidget.configure(font=(_used_font_family, MB_FontSize.TEXT.value))
    if mask:
        entryWidget.configure(show=mask)
    entryWidget.pack(side=tk.LEFT, padx="3m")
    entryWidget.bind("<Return>", __enterboxGetText)
    entryWidget.bind("<Escape>", __enterboxCancel)

    # put text into the entryWidget and have it pre-highlighted
    if __enterboxDefaultText != "":
        entryWidget.insert(0, __enterboxDefaultText)
        entryWidget.select_range(0, tk.END)

    # ------------------ ok button -------------------------------
    okButton = tk.Button(buttonsFrame, takefocus=1, text=MB_ButtonType.OK)
    _bindArrows(okButton)
    okButton.pack(expand=1, side=tk.LEFT, padx="3m", pady="3m", ipadx="2m", ipady="1m")

    # for the commandButton, bind activation events to the activation event handler
    commandButton = okButton
    handler = __enterboxGetText
    for selectionEvent in STANDARD_SELECTION_EVENTS:
        commandButton.bind("<%s>" % selectionEvent, handler)

    # ------------------ cancel button -------------------------------
    cancelButton = tk.Button(buttonsFrame, takefocus=1, text=MB_ButtonType.CANCEL)
    _bindArrows(cancelButton)
    cancelButton.pack(
        expand=1, side=tk.RIGHT, padx="3m", pady="3m", ipadx="2m", ipady="1m"
    )

    # for the commandButton, bind activation events to the activation event handler
    commandButton = cancelButton
    handler = __enterboxCancel
    for selectionEvent in STANDARD_SELECTION_EVENTS:
        commandButton.bind("<%s>" % selectionEvent, handler)

    # ------------------- time for action! -----------------
    entryWidget.focus_force()  # put the focus on the entryWidget
    boxRoot.attributes("-topmost", True)     # agd
    boxRoot.deiconify()
    _timeout_id: str = None
    if timeout is not None:
        _timeout_id = boxRoot.after(timeout, timeoutBoxRoot)
        # boxRoot.after(timeout, )
    boxRoot.mainloop()  # run it!

    # -------- after the run has completed ----------------------------------
    if _timeout_id is not None:
        boxRoot.after_cancel(_timeout_id)

    if root:
        root.deiconify()
    try:
        boxRoot.destroy()  # button_click didn't destroy boxRoot, so we do it now
    except tk.TclError:
        if __enterboxText != TIMEOUT_RETURN_VALUE:
            return None

    return __enterboxText


def __enterboxGetText(event):
    global __enterboxText

    __enterboxText = entryWidget.get()
    boxRoot.quit()


def __enterboxRestore(event):
    global entryWidget

    entryWidget.delete(0, len(entryWidget.get()))
    entryWidget.insert(0, __enterboxDefaultText)


def __enterboxCancel(event):
    global __enterboxText

    __enterboxText = None
    boxRoot.quit()

if __name__ == "__main__":
    alert('This is an alert box', 'ALERT1')
    _used_font_family = MB_FontFamily.MONOSPACE
    _used_font_size = MB_FontSize.MONOSPACE.value

    # MSGBOX_WIDTH = 600
    txt = ''
    for k,v in tk.__dict__.items():
        if not k.startswith('_') and isinstance(v, int):
            txt += f'{k:20} {v}\n'
    
    alert(txt,"ALERT2-MULTILINE")

    confirm('this is a confirm box, no timeout', "CONFIRM")
    confirm('this is a confirm box, 5 sec timeout', "CONFIRM", timeout=5000)
    prompt('This is a prompt box', 'PROMPT', 'default')
    password('This is a password box', 'PASSWORD', 'default')



"""
This module will demo the dt_tools.console.spinner capabilities including:

- Spinner with and without elapsed time
- Showcase different spinner types

To run this demo standalone:
    `poetry run python -m dt_tools.cli.dt_spinner_demo`

"""
import time

import dt_tools.logger.logging_helper as lh
from dt_tools.console.console_helper import ColorFG, TextStyle
from dt_tools.console.console_helper import ConsoleHelper as console
from dt_tools.console.spinner import Spinner, SpinnerType
from dt_tools.os.os_helper import OSHelper

def demo():    
    OSHelper.enable_ctrl_c_handler()
    console.print('')
    console.print_line_separator('Spinner Demo', 40)
    console.print('')
    
    console.print(console.cwrap("Basic code example:", style=TextStyle.ITALIC))
    console.print(console.cwrap("  spinner = Spinner(caption='Generating... ' , spinner=SpinnerType.NORMAL_SPINNER, show_elapsed=True)", style=TextStyle.ITALIC))
    console.print(console.cwrap("  spinner.start('begin work')", style=TextStyle.ITALIC))
    console.print(console.cwrap("     do some work...", style=TextStyle.ITALIC))
    console.print(console.cwrap("  spinner.caption_suffix('still working...)", style=TextStyle.ITALIC))
    console.print(console.cwrap("     do some more work...", style=TextStyle.ITALIC))
    console.print(console.cwrap("  spinner.stop_spinner()", style=TextStyle.ITALIC))
    console.print('')

    console.print(console.cwrap('Spinner, no elapsed time or status information:', style=TextStyle.UNDERLINE))
    spinner = Spinner(caption='Basic Spinner', spinner=SpinnerType.NORMAL_SPINNER)
    spinner.start_spinner()
    time.sleep(10)
    spinner.stop_spinner()
    console.clear_line(-1)

    console.print(console.cwrap('Spinner, with elapsed time:', style=TextStyle.UNDERLINE))
    spinner = Spinner(caption='Basic Spinner w/Elapsed time', spinner=SpinnerType.NORMAL_SPINNER, show_elapsed=True)
    spinner.start_spinner()
    time.sleep(10)
    spinner.stop_spinner()
    console.clear_line(-1)

    console.print(console.cwrap('Loop thru all spinner types, with elapsed time', style=TextStyle.UNDERLINE))
    sleep_time = .25
    max_range = 25
    for spinner_type in SpinnerType:
        spinner = Spinner(caption=spinner_type, spinner=spinner_type, show_elapsed=True)
        spinner.start_spinner()
        for cnt in range(1, max_range + 1):
            if cnt % 5 == 0:
                spinner.caption_suffix(f'working... {cnt} of {max_range}')
            time.sleep(sleep_time)
        spinner.stop_spinner()
    console.clear_line(-1)
    
    console.print('')
    console.print(f"End of {console.cwrap('Spinner',ColorFG.YELLOW)} demo.")

    input('\nPress Enter to continue...')

if __name__ == '__main__':
    lh.configure_logger(log_level='INFO')
    demo()


"""
Console utilities for controlling terminal screen.


Package contains two main classes for working with console windows and text.

- **ConsoleHelper**: Class to aid in console window control and text output including
    - writing to specific locations.
    - clearing portions of the screen
    - colorized output

- **ConsoleInputHelper**: Class providing input prompt with:
    - Input field editing (i.e. controlling valid input)
    - Wait time, timeout before default response is returned


Additionally, helper classes/namespaces provided:

- **ConsoleColor**: Color codes for ansi output 
    (see :func:`~dt_tools.console.console_helper.ConsoleHelper.cwrap()` function).
- **CursorShape**: Ansi codes for controlling cursor shape.

"""
import os
import re
import signal
import sys
import time
from enum import Enum
from typing import Final, List, Tuple, Union
import threading

from loguru import logger as LOGGER

from dt_tools.misc.helpers import StringHelper
from dt_tools.os.os_helper import OSHelper

# LOCK = threading.Lock()

if OSHelper.is_windows():
    import msvcrt
    from ctypes import byref, windll, wintypes  # noqa: F401
else:
    import termios
    import tty

# TODO:
#   update _output_to_terminal to allow for printing to stderr OR stdout (default)

class _ConsoleControl:
    """Console control characters."""
    ESC: Final = "\x1B"
    BELL: Final = '\a'
    CEND: Final = f'{ESC}[0m'

class ColorFG:
    """ Console font colors to be used with :func:`~dt_tools.console.console_helper.ConsoleHelper.cwrap()`."""
    DEFAULT: Final = f'{_ConsoleControl.ESC}[39m'
    BLACK: Final  = f'{_ConsoleControl.ESC}[30m'
    RED: Final    = f'{_ConsoleControl.ESC}[31m'
    GREEN: Final  = f'{_ConsoleControl.ESC}[32m'
    YELLOW: Final = f'{_ConsoleControl.ESC}[33m'
    BLUE: Final   = f'{_ConsoleControl.ESC}[34m'
    VIOLET: Final = f'{_ConsoleControl.ESC}[35m'
    BEIGE: Final  = f'{_ConsoleControl.ESC}[36m'
    WHITE: Final = f'{_ConsoleControl.ESC}[37m'

    GREY: Final    = f'{_ConsoleControl.ESC}[90m'
    RED2: Final    = f'{_ConsoleControl.ESC}[91m'
    GREEN2: Final  = f'{_ConsoleControl.ESC}[92m'
    YELLOW2: Final = f'{_ConsoleControl.ESC}[93m'
    BLUE2: Final   = f'{_ConsoleControl.ESC}[94m'
    VIOLET2: Final = f'{_ConsoleControl.ESC}[95m'
    BEIGE2: Final  = f'{_ConsoleControl.ESC}[96m'
    WHITE2: Final  = f'{_ConsoleControl.ESC}[97m'

class ColorBG:
    """Console background font colors to be used with :func:`~dt_tools.console.console_helper.ConsoleHelper.cwrap()`."""
    DEFAULT: Final = f'{_ConsoleControl.ESC}[49m'
    BLACK: Final   = f'{_ConsoleControl.ESC}[40m'
    RED: Final     = f'{_ConsoleControl.ESC}[41m'
    GREEN: Final   = f'{_ConsoleControl.ESC}[42m'
    YELLOW: Final  = f'{_ConsoleControl.ESC}[43m'
    BLUE: Final    = f'{_ConsoleControl.ESC}[44m'
    VIOLET: Final  = f'{_ConsoleControl.ESC}[45m'
    BEIGE: Final   = f'{_ConsoleControl.ESC}[46m'
    WHITE: Final   = f'{_ConsoleControl.ESC}[47m'

    GREY: Final    = f'{_ConsoleControl.ESC}[100m'
    RED2: Final    = f'{_ConsoleControl.ESC}[101m'
    GREEN2: Final  = f'{_ConsoleControl.ESC}[102m'
    YELLOW2: Final = f'{_ConsoleControl.ESC}[103m'
    BLUE2: Final   = f'{_ConsoleControl.ESC}[104m'
    VIOLET2: Final = f'{_ConsoleControl.ESC}[105m'
    BEIGE2: Final  = f'{_ConsoleControl.ESC}[106m'
    WHITE2: Final  = f'{_ConsoleControl.ESC}[107m'

class TextStyle:
    """Constants for formatting strings."""
    TRANSPARENT: Final = f'{_ConsoleControl.ESC}[0m' 
    RESET: Final       = f'{_ConsoleControl.ESC}[0m'
    BOLD: Final        = f'{_ConsoleControl.ESC}[1m'
    DIM: Final         = f'{_ConsoleControl.ESC}[2m]'
    ITALIC: Final      = f'{_ConsoleControl.ESC}[3m'
    UNDERLINE: Final   = f'{_ConsoleControl.ESC}[4m'
    BLINK: Final       = f'{_ConsoleControl.ESC}[5m'
    BLINK2: Final      = f'{_ConsoleControl.ESC}[6m'
    SELECTED: Final    = f'{_ConsoleControl.ESC}[7m'
    HIDDEN: Final      = f'{_ConsoleControl.ESC}[8m'
    STRIKETHRU: Final  = f'{_ConsoleControl.ESC}[9m'
    INVERSE: Final     = f'{_ConsoleControl.ESC}[k'
    SPACER: Final      = ' ͏'

class CursorShape(Enum):
    """Constants defining available cursor shapes."""
    DEFAULT             = f'{_ConsoleControl.ESC}[0 q'
    BLINKING_BLOCK      = f'{_ConsoleControl.ESC}[1 q'
    STEADY_BLOCK        = f'{_ConsoleControl.ESC}[2 q'
    BLINKING_UNDERLINE  = f'{_ConsoleControl.ESC}[3 q'
    STEADY_UNDERLINE    = f'{_ConsoleControl.ESC}[4 q'
    BLINKING_BAR        = f'{_ConsoleControl.ESC}[5 q'
    STEADY_BAR          = f'{_ConsoleControl.ESC}[6 q'

class _CursorAttribute(Enum):
    """Cursor control characters (Visability, Blink?)."""
    HIDE = f'{_ConsoleControl.ESC}[?25l'
    SHOW = f'{_ConsoleControl.ESC}[?25h'

class _CursorClear:
    EOS: Final = f'{_ConsoleControl.ESC}[0J'
    """Clear to End-Of-Screen."""
    BOS: Final = f'{_ConsoleControl.ESC}[1J'
    """Clear to Beginning-Of-Screen."""
    LINE: Final = f'{_ConsoleControl.ESC}[2K'
    """Clear current line."""
    EOL: Final = f'{_ConsoleControl.ESC}[0K'
    """Clear from current position to End-Of-Line."""
    BOL: Final = f'{_ConsoleControl.ESC}[1K'
    """Clear from current position to Beginning-Of-Line."""
    SCREEN: Final = f'{_ConsoleControl.ESC}[2J'
    """Clear entire screen."""

class WindowControl:
    """Window control characters (Hide/Show, Title)."""
    WINDOW_HIDE: Final  = f'{_ConsoleControl.ESC}[2t'
    WINDOW_SHOW: Final  = f'{_ConsoleControl.ESC}[1t'
    WINDOW_TITLE: Final = f'{_ConsoleControl.ESC}]2;%title%\a'


# https://docs.microsoft.com/en-us/windows/console/console-virtual-terminal-sequences
# https://www.lihaoyi.com/post/BuildyourownCommandLinewithANSIescapecodes.html
# https://invisible-island.net/xterm/ctlseqs/ctlseqs.html

# ==========================================================================================================
class ConsoleHelper():
    """
    Class to assist with console output.  
    
    Methods to:

        - Set cursor shape and visibility.
        - Set console window title.
        - Clear functions.
        - Cursor control (up, down, left, right, move to location,...).

    Example::
        from dt_tools.console.console_helper import ConsoleHelper, ConsoleColors

        con = ConsoleHelper()
        con.clear_screen(cursor_home=True)

        console_size = con.get_console_size()
        row, col = con.cursor_current_position()
        con.print_at(5,5, f'Console size: {console_size}, cur pos: {row},{col}', eol='\\n')
        
        token = f'Test {con.cwrap("Yellow", ConsoleColors.CYELLOW)} string')
        con.print_at(5, 10, token)
        
    """
    LAST_CONSOLE_STR: str = None
    
    @classmethod
    def cursor_set_attribute(cls, attr: Union[_CursorAttribute, str]):
        token = attr.value if isinstance(attr, _CursorAttribute) else attr
        cls._output_to_terminal(token)

    @classmethod
    def cursor_set_shape(cls, shape: Union[CursorShape, str]):
        token = shape.value if isinstance(shape, CursorShape) else shape
        cls._output_to_terminal(token)

    @classmethod
    def get_console_size(cls) -> Tuple[int, int]:
        """
        Return console size in rows and columns.

        Returns:
            Size as (rows, columns).
        """
        try:
            size = os.get_terminal_size()
            rows = int(size.lines)
            columns = int(size.columns)
        except OSError:
            rows = 0
            columns = 0
        return (rows, columns)

    @classmethod
    def valid_console(cls) -> bool:
        try:
            _ = os.get_terminal_size()
            return True
        except OSError:
            return False
        
    @classmethod
    def console_hide(cls):
        """Minimize console/terminal window"""
        cls._output_to_terminal(WindowControl.WINDOW_HIDE)
    
    @classmethod
    def console_show(cls):
        """Restore console/terminal window"""
        cls._output_to_terminal(WindowControl.WINDOW_SHOW)
        
    @classmethod
    def set_console_viewport(cls, start_row: int = None, end_row: int = None):
        """
        Set console scrollable area to start_row / end_row.  Default is whole screen.

        The viewport defines area (rows) text scrolls within.  If no
        arguments provided, viewport is defaulted to whole screen.

        Keyword Arguments:
            start_row: Staring row of viewport (default: {None}).
            end_row: Ending row of viewport (default: {None}).

        Raises:
            ValueError: Invalid start|end row.
        """

        max_row, max_col = cls.get_console_size()
        starting_row = 1 if start_row is None else start_row
        ending_row = int(max_row) if end_row is None else end_row
        if starting_row < 1 or starting_row > ending_row or starting_row > max_row:
            raise ValueError(f"set_viewport(): Invalid start row: {start_row}")
        if ending_row > max_row:
            raise ValueError(f'set_viewport(): Invalid end row: {end_row}')

        cls._output_to_terminal(f'{_ConsoleControl.ESC}[{starting_row};{ending_row}r')

    @classmethod
    def console_title(cls, title: str):
        """
        Set the console/window title.

        Arguments:
            title: String to be displayed on the title bar.
        """
        title_cmd = WindowControl.WINDOW_TITLE.replace("%title%", title)
        print(title_cmd)
        cls._output_to_terminal(title_cmd)

    @classmethod
    def cursor_current_position(cls) -> Tuple[int, int]:
        """
        Current cursor location.

        Returns:
            Cusor location: (row, col).
        """
        if OSHelper.is_windows():
            return cls._get_windows_cursor_position()
        
        return cls._get_linux_cursor_position()
            

    @classmethod
    def cursor_save_position(cls):
        """
        Save cursor position, can be restored with restore_position() call.
        """
        cls._output_to_terminal(f'{_ConsoleControl.ESC}[s')

    @classmethod
    def cursor_restore_position(cls):
        """
        Restore cursor position, saved with the save_position() call.
        """
        cls._output_to_terminal(f'{_ConsoleControl.ESC}[u')

    @classmethod
    def cursor_off(cls):
        cls._output_to_terminal(_CursorAttribute.HIDE.value)

    @classmethod
    def cursor_on(cls):
        """Turn console cursor on"""
        cls._output_to_terminal(_CursorAttribute.SHOW.value)

    @classmethod
    def clear_screen(cls, cursor_home: bool = True):
        """
        Clear screen and home cursor.

        Keyword Arguments:
            cursor_home: If true home cursor else leave at current position (default: {True}).
        """
        cls._output_to_terminal(_CursorClear.SCREEN)
        if cursor_home:
            cls.cursor_move(1, 1)

    @classmethod
    def clear_to_EOS(cls):
        """Clear from cursor to end of screen"""
        cls._output_to_terminal(_CursorClear.EOS)

    @classmethod
    def clear_to_BOS(cls):
        """Clear from cursor to beginning of screen"""
        cls._output_to_terminal(_CursorClear.BOS)

    @classmethod
    def clear_line(cls, row_offset: int = 0):
        """
        Clear console line, and position cursor at beginning of line.

        Args:
            row_offset (int, optional): Offset of line to clear. Defaults to 0.
        """
        if row_offset < 0:
            cls.cursor_up(abs(row_offset))
        elif row_offset > 0:
            cls.cursor_down(row_offset)
        cls._output_to_terminal(_CursorClear.LINE)
        cls.cursor_move(-1, 1)

    @classmethod
    def clear_to_EOL(cls):
        """Clear from cursor to end of line"""
        cls._output_to_terminal(_CursorClear.EOL)
    
    @classmethod
    def clear_to_BOL(cls):
        """Clear from cursor to beginning of line"""
        cls._output_to_terminal(_CursorClear.BOL)

    @classmethod
    def cursor_up(cls, steps: int = 1):
        """
        Move cursor up.

        Keyword Arguments:
            steps: Number of rows to move up (default: {1}).
        """
        cls._output_to_terminal(f'{_ConsoleControl.ESC}[{steps}A')

    @classmethod
    def cursor_down(cls, steps: int = 1):
        """
        Move cursor down.

        Keyword Arguments:
            steps: Number of rows to move down (default: {1}).
        """
        cls._output_to_terminal(f'{_ConsoleControl.ESC}[{steps}B')

    @classmethod
    def cursor_right(cls, steps: int = 1):
        """
        Move cursor right.

        Keyword Arguments:
            steps: Number of columns to move right (default: {1}).
        """
        cls._output_to_terminal(f'{_ConsoleControl.ESC}[{steps}C')

    @classmethod
    def cursor_left(cls, steps: int = 1):
        """
        Move cursor left.

        Keyword Arguments:
            steps: Number of columns to move left (default: {1}).
        """
        cls._output_to_terminal(f'{_ConsoleControl.ESC}[{steps}D')

    @classmethod
    def cursor_scroll_up(cls, steps: int = 1):
        """
        Scroll screen contents up.

        Keyword Arguments:
            steps: Number of row to scroll up (default: {1}).
        """
        cls._output_to_terminal(f'{_ConsoleControl.ESC}[{steps}S')

    @classmethod
    def cursor_scroll_down(cls, steps: int = 1):
        """
        Scroll screen contents down.

        Keyword Arguments:
            steps: Number of row to scroll down (default: {1}).
        """
        cls._output_to_terminal(f'{_ConsoleControl.ESC}[{steps}T')

    @classmethod
    def cursor_move(cls, row:int = -1, column:int = -1) -> bool:
        """
        Move cursor to spefic location on console.

        If row or column is not set, current position (row or column) will be used.

        Keyword Arguments:
            row: Row to move cursor (default: {-1}).
            column: Column to move cursor (default: {-1}).

        Returns:
            True if successful, False if location not valid.
        """

        cur_row, cur_col = cls.cursor_current_position()
        if  row <= 0:
            row = int(cur_row)
        if column <= 0:
            column = int(cur_col)
        max_rows, max_columns = cls.get_console_size()
        if row <= 0 or column <= 0:
            LOGGER.debug('cursor_move - row/column must be > 0')
            return False
        if max_rows > 0 and max_columns > 0:
            if column > max_columns or row > max_rows:
                LOGGER.debug((f'cursor_move - row > {max_rows} or col > {max_columns}'))
                return False
        
        cls._output_to_terminal(f"{_ConsoleControl.ESC}[%d;%dH" % (row, column))    
        return True

    @classmethod
    def print(cls, msg, eol='\n', as_bytes:bool = False, to_stderr:bool = False, 
              fg: ColorFG = ColorFG.DEFAULT, bg: ColorBG = ColorBG.DEFAULT, style: TextStyle = TextStyle.TRANSPARENT): # type: ignore
        """
        Print msg to console.

        Arguments:  
            msg (any): Message to print to console.  
            eol (str, optional): End of line character. Defaults to '\\n'.  
            as_bytes (bool, optional): Output as bytes (ie. raw string). Defaults to False.  
            to_stderr (bool, optional): Print to stderr (instead of stdout). Defaults to False.  

        """
        out_msg = cls.cwrap(msg, fg=fg, bg=bg, style=style)
        cls._output_to_terminal(out_msg, eol=eol, as_bytes=as_bytes, to_stderr=to_stderr)

    @classmethod
    def print_at(cls, row: int, col: int, text: str, eol='', as_bytes: bool = False, to_stderr:bool = False) -> bool:
        """
        Print text at specific location on console.

        No new-line will occur unless eol parameter is specifed as '\\n'.

        Arguments:
            row: Target row.
            col: Target column.
            text: String to write to console.

        Keyword Arguments:
            eol: End of line character (default: {''}).
            as_bytes: Output msg as bytes (default: {False}).
            to_stderr: Output to stderr (instead of stdout) (default: {False}).

        Returns:
            True if print to console successful, False if invalid location.
        """
        if cls.cursor_move(row, col):
            cls._output_to_terminal(text, eol=eol, as_bytes=as_bytes, to_stderr=to_stderr)
            return True
        return False
        
    @classmethod
    def print_with_wait(cls, text: str, wait: float = 0.0, eol='\n'):
        """
        Print text at current location and wait specified number of seconds.

        Arguments:
            text: Text to be printed.

        Keyword Arguments:
            wait: Number of seconds to wait (default: {0.0}).
            eol: EndOfLine character (default: {'\\n'}).

        """
        print(text, end=eol, flush=True)
        if wait > 0:
            time.sleep(wait)

    @classmethod
    def display_status(cls, text, wait: int = 0, status_eyecatcher: bool = True):
        """
        Display status message on last row of screen.

        Arguments:
            text: Status message to be displayed.

        Keyword Arguments:
            wait: Number of seconds to wait/pause (default: {0}).
        """
        max_row, max_col = cls.get_console_size()
    
        save_row, save_col = cls.cursor_current_position()
        if status_eyecatcher:
            eyecatcher_style = f'{ColorBG.GREY}{ColorFG.WHITE2}'
            # inverse_token = f'{eyecatcher_style}{TextStyle.INVERSE}'
            text = f'{eyecatcher_style}{text.replace(TextStyle.RESET, f"{TextStyle.RESET}{eyecatcher_style}")}'
            # cls.print(text, as_bytes=True)
            
        cls.print_at(max_row, 1, f'{text}', eol='')     
        cls.clear_to_EOL()
        cls.print_at(1, 1, TextStyle.RESET)   
        cls.cursor_move(save_row, save_col)
        if wait > 0:
            time.sleep(wait)
    
    @classmethod
    def print_line_separator(cls, text: str = '', length: int = -1):
        """
        Print line separator at current cursor position.

        Keyword Arguments:
            text: Text to be displayed within the separator line (default: {''}).
            length: Lenght of the separator line  (default: {-1}).
            if < 0, use console width.
        """
        print(cls.sprint_line_separator(text, length))

    @classmethod
    def sprint_line_separator(cls, text: str = '', length: int = -1) -> str:
        """
        Return string underline (separator) with optional text.

        Keyword Arguments:
            text: Text to be displayed within the separator line (default: {''}).
            length: Length of the separator line  (default: {-1}).
            if < 0, use console width.

        Returns:
            Separator line string.
        """
        if length < 0:
            row, col = cls.cursor_current_position()
            max_rows, max_cols = cls.get_console_size()
            length = max_cols - col
        fill_len = length - len(cls.remove_nonprintable_characters(text))
        if TextStyle.RESET in text:
            color_code = cls.color_code(style=TextStyle.UNDERLINE, fg=ColorFG.DEFAULT, bg=ColorBG.DEFAULT)
            text = text.replace(TextStyle.RESET, f'{TextStyle.RESET}{color_code}')
        # cls.print(text, as_bytes=True)
        line_out = f'{TextStyle.UNDERLINE}{text}{" "*(fill_len-1)}{TextStyle.RESET}'
        return line_out

    @classmethod
    def remove_nonprintable_characters(cls, text: str) -> str:
        """
        Return the length of strings printable characters

        Args:
            text (str): Input string

        Returns:
            int: Number of printable characters
        """
        ansi_escape = re.compile(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])')
        result = ansi_escape.sub('', text)    
        return result
    
    @classmethod
    def debug_display_cursor_location(cls, msg:str = ""):
        """
        Display current location (row, col) in status area.

        Keyword Arguments:
            msg: Message text to append after current location (default: {""}).
        """
        cls.display_status(f'Cursor: {str(ConsoleHelper().cursor_current_position())}  {msg}')

    @classmethod
    def cwrap(cls, text: str, fg: ColorFG = None, bg: ColorBG = None, style: Union[List[TextStyle],TextStyle] = None, length: int = -1) -> str:
        """ 
        Wrap text with color codes for console display.
        
        See ConsoleFG, ConsoleBG and ConsoleStyle for control codes

        Arguments:
            **text**: req - String containing text to be colorized.
            **fg**: req -  The FG color OR color string (see ColorFG)
            **bg**: opt - The BG color (see ColorBG)
            **style**: opt - The style to be applied (see TextStyleControls)
            **length**" opt - Length of string. Pad right with spaces, -1 = len(text).

        Returns:
            Updated string.
        """
        w_text = str(text)
        if length < 0:
            length = len(w_text)

        color_code = ''
        if fg and bg and style:
            color_code = ConsoleHelper.color_code(style, fg, bg)
        else:
            if fg:
                color_code += fg
            if bg:
                color_code += bg
            if style:
                if isinstance(style, list):
                   style = ''.join(style) 
                color_code += style

        padded_str = StringHelper.pad_r(w_text, length)
        ret_str =  f'{color_code}{padded_str}{_ConsoleControl.CEND}'
        # cls._output_to_terminal(ret_str, eol='\n', as_bytes=True)
        return ret_str
    

    @classmethod
    def color_code(cls, style: Union[List[TextStyle], TextStyle] = TextStyle.TRANSPARENT, fg: ColorFG = None, bg: ColorBG= None) -> str:
        """
        Create ANSI color code for style, fg color and bg color

        If any parameter (style, fg or bg) is None, current value will be used.

        Args:
            
            **style**: (:class:`~TextStyleControls`, optional): Font style (ie. bold, italic,...).  
            **fg**: (:class:`~ColorFG`, optional): Foreground text color.  
            **bg**: (:class:`~ColorBG`, optional): Background color.  

        Returns:
            str: The ANSI code representing the desired ANSI atributes.
            
        """
        codes = []
        if isinstance(style, list):
            style = ''.join(style) # type: ignore
        codes = [str(style), str(fg), str(bg)]
        format = ';'.join([x.removeprefix(f'{_ConsoleControl.ESC}[').removesuffix('m') for x in codes if x != 'None'])
        code = f'{_ConsoleControl.ESC}[{format}m'
        return code

    # == Private Function ================================================================================= 
    @classmethod   
    def _output_to_terminal(cls, token: str, eol:str='', as_bytes: bool = False, to_stderr: bool = False):
    
        output_str = bytes(token,'utf-8') if as_bytes else token
        if to_stderr:
            print(output_str, end=eol, flush=True, file=sys.stderr)
        else:
            try:
                print(output_str, end=eol, flush=True)
            except UnicodeEncodeError:
                # stderr will escape non-printable characters
                print(output_str, end=eol, flush=True, file=sys.stderr)
                
        cls.LAST_CONSOLE_STR = token

    @classmethod
    def _display_color_palette(cls):
        """
        prints table of formatted text format options
        """
        for style in range(8):
            for fg in range(30, 38):
                token = ''
                for bg in range(40, 48):
                    format = ';'.join([str(style), str(fg), str(bg)])
                    token += '\x1b[%sm %s \x1b[0m' % (format, format)
                print(token)
            print('\n')

    @classmethod
    def _get_windows_cursor_position(cls) -> Tuple[int, int]:
        valid_coords = False
        val_errcnt = 0
        row: int = -1
        col: int = -1
        while not valid_coords and val_errcnt < 3:
            sys.stdout.write("\x1b[6n")
            sys.stdout.flush()
            buffer = bytes()
            while msvcrt.kbhit():
                buffer += msvcrt.getch()
            hex_loc = buffer.decode().replace('\x1b[','').replace('R','')
            # print(hex_loc)
            token = hex_loc.split(';')
            try:
                row = int(token[0])
                col = int(token[1])
                valid_coords = True
            except ValueError as ve:
                val_errcnt += 1
                if val_errcnt == 3:
                    LOGGER.debug(f'cursor_current_postion()-Invalid row/col (hex_loc): {hex_loc} | {token} - {ve}')
        return row, col
    
    @classmethod
    def _get_linux_cursor_position(cls) -> Tuple[int, int]:
        """Gets the current cursor position on the terminal."""

        fd = sys.stdin.fileno()
        old_settings = termios.tcgetattr(fd)

        try:
            tty.setraw(sys.stdin.fileno())
            sys.stdout.write("\x1b[6n")
            sys.stdout.flush()

            response = ""
            while True:
                char = sys.stdin.read(1)
                if char == "R":
                    break
                response += char

            row, col = map(int, response[2:].split(";"))
            return row, col

        finally:
            termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)        



# ==========================================================================================================
class ConsoleInputHelper():
    """
    Helper for getting input from the console.

    Example::

        from dt_tools.console.console_helper import ConsoleInputHelper

        ih = ConsoleInputHelper()
        resp = ih.get_input_with_timeout(prompt='Pick a color > ', 
            valid_responses=['Red', 'Blue', 'Green'], 
            default='Blue', 
            timeout_secs=5)
        print(f'You selected: {resp}')    

    """

    YES_NO_RESPONSE: Final[List[str]] = ['Y','y', 'N', 'n']
    """Yes/No valid_argument list constant"""

    @classmethod
    def get_input_with_timeout(cls, prompt: str, valid_responses: list = [],  
                               default: str = None, timeout_secs: int = -1, 
                               parms_ok: bool = False) -> Union[str, Tuple[str, list]]:
        """
        Display a prompt for use input.  
        
        If valid_responses is supplied:
            - input will be validated.  User will be re-prompted on bad input. 
            - default response will be returned after timeout (if specified).

        Arguments:
            **prompt**: Text to be displayed as prompt.
            **valid_responses**: A list of valid responses (default: {[]}).
              User input must match one of the values. 
              If list is empty, all input will be accepted.
            **default**: Default value to be return on timeout (default: {None}).
            **timeout_secs**: Number of seconds to wait for response (default: {-1}).  
              If <0, no timeout, wait until user presses enter.
            **parms_ok**: Allow extra parameter input (default: {False}).
              If True, allows user to provide additional text after the valid response.

        Returns:
            User input or default value (if timeout).

        Example::

            from dt_tools.console.console_helper input ConsoleInputHelper

            ih = ConsoleInputHelper()
            resp = ih.get_input_with_timeout(prompt='Pick a color > ', 
                valid_responses=['Red', 'Blue', 'Green'], 
                default='Blue', 
                timeout_secs=5)
            print(f'You selected: {resp}')

        """
        response: str = ''
        chk_response = ''
        response_params: List = None
        valid_input = False
        while not valid_input:
            if timeout_secs < 0:
                response = input(prompt)
            else:
                try:
                    if OSHelper.is_windows():
                        response = cls._input_with_timeout_win(prompt, timeout_secs, default)
                    else:
                        response = cls._input_with_timeout_nix(prompt, timeout_secs, default)
                except TimeoutError:
                    response = default
                    valid_input = True
            
            if not parms_ok:
                chk_response = response
                response_params = None
            else:
                token = response.split()
                if len(token) > 0:
                    chk_response = token[0]
                    response_params = token[1:]
                else:
                    chk_response = response
                    response_params = None

            if not valid_responses:
                LOGGER.trace('no valid responses to check')
                valid_input = True
            elif chk_response in valid_responses:
                    valid_input = True

        if parms_ok:
            return chk_response, response_params
        
        return chk_response

    @classmethod
    def wait_with_bypass(cls, secs: int):
        """
        Pause execution for specified number of seconds.
        
        User may press enter to resume prior to timeout seconds.

        Arguments:
            secs: Number of seconds to wait.
        """
        cls.get_input_with_timeout("", timeout_secs=secs)

    @classmethod
    def _input_with_timeout_nix(cls,prompt: str, timeout_secs: int, default: str) -> str:
        # set signal handler for *nix systems
        LOGGER.trace("_input_with_timeout_nix()")
        signal.signal(signal.SIGALRM, ConsoleInputHelper._alarm_handler)
        signal.alarm(timeout_secs) # produce SIGALRM in `timeout` seconds

        response = default
        try:
            response = input(prompt)
        finally:
            signal.alarm(0) # cancel alarm
            return response

    @classmethod
    def _input_with_timeout_win(cls, prompt: str, timeout_secs: int,  default: str= None) -> str:
        LOGGER.trace("_input_with_timeout_win()")
        sys.stdout.write(prompt)
        sys.stdout.flush()
        timer = time.monotonic
        endtime = timer() + timeout_secs
        result = []
        while timer() < endtime:
            if msvcrt.kbhit():
                result.append(msvcrt.getwche()) #XXX can it block on multibyte characters?
                endtime = timer() + timeout_secs  # Reset timer each time a key is pressed.
                if result[-1] == '\r':   #XXX check what Windows returns here
                    print('')
                    return ''.join(result[:-1])
            time.sleep(0.04) # just to yield to other processes/threads
        if result:
            print('')
            return ''.join(result)
        elif default:
            print(default)
        raise TimeoutError('Time Expired.')

    def _alarm_handler(signum, frame):
        raise TimeoutError('time expired.')



if __name__ == "__main__":
    from dt_tools.cli.demos.dt_console_demo import console_helper_demo as demo
    from dt_tools.console.console_helper import ColorFG, ConsoleHelper
    demo()
    # print("ConsoleHelper.cwrap('This is a test', style=TextStyle.UNDERLINE, length=30)")
    # token = ConsoleHelper.cwrap('This is a test', style=TextStyle.UNDERLINE, length=30)
    # token2 = ConsoleHelper.remove_nonprintable_characters(token)
    # print(token)
    print("That's all folks!")


"""
This module will demo the dt_tools.console.progress_bar capabilities including:

- Display progress bar 
- Display progress bar with elapsed time


To run this demo standalone:
    `poetry run python -m dt_tools.cli.dt_progress_bar_demo`

"""
from dt_tools.console.console_helper import ColorFG, TextStyle
from dt_tools.console.console_helper import ConsoleHelper as console
from dt_tools.console.progress_bar import ProgressBar
import time

def demo():    
    console.print('')
    console.print_line_separator('ProgressBar Demo', 40)
    console.print('')

    console.print(console.cwrap('Plain progress bar', style=TextStyle.BOLD))
    console.print(console.cwrap('code:', style=[TextStyle.ITALIC]))
    console.print(console.cwrap('  pbar = ProgressBar("Progress bar", bar_length=40, max_increments=50, show_elapsed=False)', style=[TextStyle.ITALIC]))
    console.print(console.cwrap('  for incr in range(1,51):', style=[TextStyle.ITALIC]))
    console.print(console.cwrap('      pbar.display_progress(incr, f"incr [{incr}]")', style=[TextStyle.ITALIC]))
    console.print(console.cwrap('      time.sleep(sleep_time)\n', style=[TextStyle.ITALIC]))

    sleep_time = .15
    pbar = ProgressBar("Progress bar", bar_length=40, max_increments=50, show_elapsed=False)
    for incr in range(1,51):
        pbar.display_progress(incr, f'incr [{incr}]')
        time.sleep(sleep_time)    

    console.print('')
    console.print(console.cwrap('Progress bar w/elapsed time', style=TextStyle.BOLD))
    console.print(console.cwrap('code:', style=[TextStyle.ITALIC]))
    console.print(console.cwrap('  pbar = ProgressBar("Progress bar", bar_length=40, max_increments=50, show_elapsed=True)', style=[TextStyle.ITALIC]))
    console.print(console.cwrap('  for incr in range(1,51):', style=[TextStyle.ITALIC]))
    console.print(console.cwrap('      pbar.display_progress(incr, f"incr [{incr}]")', style=[TextStyle.ITALIC]))
    console.print(console.cwrap('      time.sleep(sleep_time)\n', style=[TextStyle.ITALIC]))
    pbar = ProgressBar("Progress bar", bar_length=40, max_increments=50, show_elapsed=True)
    for incr in range(1,51):
        pbar.display_progress(incr, f'incr [{incr}]')
        time.sleep(sleep_time)
    
    console.print('')
    console.print(f"End of {console.cwrap('ProgressBar', ColorFG.YELLOW)} demo.")

    input('\nPress Enter to continue...')

if __name__ == '__main__':
    demo()


"""
This module will demo the dt_tools.console.msgbox capabilities including:

- Showing the 4 types of message boxes (Alert, Confirmation, Prompt and Password)
- Show message boxes with default timeouts
- Show multi-line boxes
- Show how to retrieve message box input, both visible and hidden (i.e. password)

To run this demo standalone:
    `poetry run python -m dt_tools.cli.dt_msgbox_demo`

"""
from dt_tools.console.console_helper import ColorFG
from dt_tools.console.console_helper import ConsoleHelper as console
from dt_tools.os.os_helper import OSHelper

import tkinter as tk

import dt_tools.console.msgbox as msgbox

def demo():
    OSHelper.enable_ctrl_c_handler()

    console.print('')
    console.print_line_separator('Alert box (no timeout)', 40)

    resp = msgbox.alert('This is an alert box', 'ALERT no timeout')
    console.print(f'  returns: {console.cwrap(resp, ColorFG.GREEN)}')

    console.print('')
    console.print_line_separator('Alert box (w/timeout, 3 sec)', 40)
    resp = msgbox.alert('This is an alert box', 'ALERT w/Timeout', timeout=3000)
    console.print(f'  returns: {console.cwrap(resp, ColorFG.GREEN)}')

    txt = ''
    for k,v in tk.__dict__.items():
        if not k.startswith('_') and isinstance(v, int):
            txt += f'{k:20} {v}\n'
    
    console.print('')
    console.print_line_separator('Alert box (multi-line)', 40)
    msgbox.set_font(msgbox.MB_FontFamily.MONOSPACE, msgbox.MB_FontSize.MONOSPACE)
    resp = msgbox.alert(txt,"ALERT-MULTILINE (no timeout)")
    console.print(f'  returns: {console.cwrap(resp, ColorFG.GREEN)}')
    
    console.print('')
    console.print_line_separator('Confirmation box (no timeout)', 40)
    msgbox.set_font(msgbox.MB_FontFamily.PROPORTIONAL, msgbox.MB_FontSize.PROPORTIONAL)
    resp = msgbox.confirm('this is a confirm box, no timeout', "CONFIRM")
    console.print(f'  returns: {console.cwrap(resp, ColorFG.GREEN)}')
    
    console.print('')
    console.print_line_separator('Confirmation box (3 sec timeout)', 40)
    resp = msgbox.confirm('this is a confirm box, 3 sec timeout', "CONFIRM", timeout=3000)
    console.print(f'  returns: {console.cwrap(resp, ColorFG.GREEN)}')
    
    console.print('')
    console.print_line_separator('Prompt box (no timeout)', 40)
    resp = msgbox.prompt('This is a prompt box', 'PROMPT', 'default')
    console.print(f'  returns: {console.cwrap(resp, ColorFG.GREEN)}')
    
    console.print('')
    console.print_line_separator('Prompt box (3 sec timeout)', 40)
    resp = msgbox.prompt('This is a prompt box', 'PROMPT (3 sec timeout)', 'default', timeout=3000)
    console.print(f'  returns: {console.cwrap(resp, ColorFG.GREEN)}')
    
    console.print('')
    console.print_line_separator('Password box (no timeout)', 40)
    resp = msgbox.password('This is a password box', 'PASSWORD', 'SuperSecretPassword')
    console.print(f'  returns: {console.cwrap(resp, ColorFG.GREEN)}')

    console.print('')
    console.print(f"End of {console.cwrap('MessageBox', ColorFG.YELLOW)} demo.")

    input('\nPress Enter to continue...')


if __name__ == '__main__':
    demo()


"""
Progress bar to provide visual feedback of process activity.

Features:
    - Customized display.
    - Optionally display elapsed time.
    - Progress bar reflects percent completion of process.

Example::

    from dt_tools.console.progress_bar import ProgressBar

    p_bar = ProgressBar("Sample Execution", bar_line=40, max_increments=100)
    for i in range(99):
        p_bar.display_progress(i+1)
        do something....
    p_bar.cancel_progress()
    
"""

import os
import time
from datetime import datetime as dt

from loguru import logger as LOGGER

from dt_tools.console.console_helper import ConsoleHelper


class ProgressBar():
    """
    Displays progress bar in the console.  
    """

    def __init__(self, caption: str, bar_length: int, max_increments: int, fill= '█', str_end = "\r", show_elapsed: bool = False, show_pct: bool = True):
        """
        Progress bar class instantiation

        Arguments:
            caption -- Caption text
            bar_length -- Length of progress bar, must be < console width
            max_increments -- Maximum number of progress segments

        Keyword Arguments:
            fill -- Progress fill character (default: {'█'})
            str_end -- End of progress bar string. default is progress remains fixed location (default: {"\r"})
            show_elapsed -- Append elapsed time at end of progress bar (default: {False})
            show_pct -- Append pct complete at end of progress bar (default: {True})
        Note:
            To update progress, call display_progress(inc) method to indicate current increment count.  
            When inc varible reaches max_increments value, progress bar will terminate.

        Example::
            from dt_tools.console.progress_bar import ProgressBar
            
            p_bar = ProgressBar("Sample Execution", bar_line=40, max_increments=100)
            for i in range(99):
                p_bar.display_progress(i+1)
                do something....
            p_bar.cancel_progress()

        """
        _, max_bar_len = ConsoleHelper.get_console_size()
        max_bar_len -= len(caption)
        if show_elapsed:
            max_bar_len -= 8  # len of elapsed str
        if show_pct:
            max_bar_len -= 5  # len of pct str

        self._caption = caption
        self._bar_len = min(max_bar_len, bar_length)
        self._max_increments = max_increments
        self._fill = fill
        self._str_end = str_end
        self._show_elapsed = show_elapsed
        self._show_pct: bool = show_pct
        
        self._start_time = dt.now()
        self._decimals = 1
        self._started = False
        self._elapsed_time = '00:00:00'
        self._finished = False
        self.console = ConsoleHelper()
        LOGGER.trace('ProgressBar initialized.')

    def display_progress(self, current_increment: int, suffix: str = ''):
        """
        Update the progress bar filling up to current increment.
        
        Parameters:
            current_increment: integer indication progress up to max_increments
            suffix:            text to display to right of progress bar, can be used to show status
        """
        if not self._started:
            self._start_time = dt.now()
            self._started = True
        if current_increment > self._max_increments:
            current_increment = self._max_increments

        self.console.cursor_off()
        self._finished = False
        term_columns = os.get_terminal_size().columns
        if ConsoleHelper.valid_console(): 
            filled_len = int(self._bar_len * current_increment // self._max_increments)
            bar = self._fill * filled_len + '-' * (self._bar_len - filled_len)
            
            display_line = f'\r{self._caption} [{bar}]'
            if self._show_pct and len(display_line) + 6 < term_columns:
                cur_percent = 100 * (current_increment / self._max_increments)
                dsply_percent = f'{cur_percent:5.1f}%'
                display_line += f' {dsply_percent}'
            
            if len(suffix) > 0 and len(display_line) + len(suffix)+1 < term_columns:
                display_line += f' {suffix}'

            self._elapsed_time = self._calculate_elapsed_time(dt.now(), self._start_time)
            if self._show_elapsed and len(display_line) + len(self.elapsed_time) + 1 < term_columns:
                display_line += f' {self._elapsed_time}'
            
            # if len(display_line) > self._term_columns:
            #     terminal_line = (display_line[:self._term_columns-7] + '...' + display_line[-3:]) 
            # else:
            #     terminal_line = display_line
            # terminal_line = (display_line[:self._term_columns-7] + '...' + display_line[-3:]) if len(display_line) > self._term_columns else display_line
            terminal_line = display_line
            print(terminal_line, end = self._str_end, flush=True)

        if current_increment >= self._max_increments:
            self.cancel_progress()

    def cancel_progress(self):
        """Turn off progress bar."""
        self._finshed = True
        self._elapsed_time = self._calculate_elapsed_time(dt.now(), self._start_time)
        self.console.cursor_on()
        self.console.print('')
        self._started = False

    @property
    def elapsed_time(self) -> str:
        """Return elapsed time in format hh:mm:ss."""
        return self._elapsed_time
    
    def _calculate_elapsed_time(self, end_time: float, start_time: float) -> str:
        diff = end_time - start_time
        hours = diff.seconds // 3600
        minutes = diff.seconds // 60 % 60
        seconds = diff.seconds % 60
        self._elapsed_time = f"{hours:02d}h:{minutes:02d}m:{seconds:02d}s"
        return self._elapsed_time


if __name__ == "__main__":
    print('')
    b_len = 80
    m_inc = 140
    pbar = ProgressBar("Test bar", bar_length=b_len, max_increments=m_inc, show_elapsed=False)
    for incr in range(1,m_inc+1):
        pbar.display_progress(incr, f'incr [{incr}]')
        time.sleep(.05)    

    print('\nProgress bar with no pct complete...')
    pbar = ProgressBar("Test bar", bar_length=b_len, max_increments=m_inc, show_elapsed=False, show_pct=False)
    for incr in range(1,m_inc+1):
        pbar.display_progress(incr, f'incr [{incr}]')
        time.sleep(.05)    

    print('\nProgress bar with elapsed time...')
    pbar = ProgressBar("Test bar", bar_length=b_len, max_increments=m_inc, show_elapsed=True)
    for incr in range(1,m_inc+1):
        pbar.display_progress(incr, f'incr [{incr}]')
        time.sleep(.05)

